<?php declare(strict_types=1);

namespace Rector\DeadCode\Rector\Class_;

use PhpParser\Node;
use PhpParser\Node\Expr\Assign;
use PhpParser\Node\Expr\New_;
use PhpParser\Node\Expr\PropertyFetch;
use PhpParser\Node\Stmt\Class_;
use PhpParser\Node\Stmt\ClassMethod;
use PhpParser\Node\Stmt\Property;
use Rector\DeadCode\Doctrine\DoctrineEntityManipulator;
use Rector\DeadCode\UnusedNodeResolver\ClassUnusedPrivateClassMethodResolver;
use Rector\NodeContainer\ParsedNodesByType;
use Rector\NodeTypeResolver\Node\AttributeKey;
use Rector\PhpParser\Node\Manipulator\ClassManipulator;
use Rector\Rector\AbstractRector;
use Rector\RectorDefinition\CodeSample;
use Rector\RectorDefinition\RectorDefinition;

/**
 * @see \Rector\DeadCode\Tests\Rector\Class_\RemoveUnusedDoctrineEntityMethodAndPropertyRector\RemoveUnusedDoctrineEntityMethodAndPropertyRectorTest
 */
final class RemoveUnusedDoctrineEntityMethodAndPropertyRector extends AbstractRector
{
    /**
     * @var ParsedNodesByType
     */
    private $parsedNodesByType;

    /**
     * @var Assign[]
     */
    private $collectionByPropertyName = [];

    /**
     * @var ClassUnusedPrivateClassMethodResolver
     */
    private $classUnusedPrivateClassMethodResolver;

    /**
     * @var ClassManipulator
     */
    private $classManipulator;

    /**
     * @var DoctrineEntityManipulator
     */
    private $doctrineEntityManipulator;

    public function __construct(
        ParsedNodesByType $parsedNodesByType,
        ClassUnusedPrivateClassMethodResolver $classUnusedPrivateClassMethodResolver,
        ClassManipulator $classManipulator,
        DoctrineEntityManipulator $doctrineEntityManipulator
    ) {
        $this->parsedNodesByType = $parsedNodesByType;
        $this->classUnusedPrivateClassMethodResolver = $classUnusedPrivateClassMethodResolver;
        $this->classManipulator = $classManipulator;
        $this->doctrineEntityManipulator = $doctrineEntityManipulator;
    }

    public function getDefinition(): RectorDefinition
    {
        return new RectorDefinition('Removes unused methods and properties from Doctrine entity classes', [
            new CodeSample(
                <<<'CODE_SAMPLE'
use Doctrine\ORM\Mapping as ORM;

/**
 * @ORM\Entity
 */
class UserEntity
{
    /**
     * @ORM\Column
     */
    private $name;

    public function getName()
    {
        return $this->name;
    }

    public function setName($name)
    {
        $this->name = $name;
    }
}
CODE_SAMPLE
                ,
                <<<'CODE_SAMPLE'
use Doctrine\ORM\Mapping as ORM;

/**
 * @ORM\Entity
 */
class UserEntity
{
}
CODE_SAMPLE
            ),
        ]);
    }

    /**
     * @return string[]
     */
    public function getNodeTypes(): array
    {
        return [Class_::class];
    }

    /**
     * @param Class_ $node
     */
    public function refactor(Node $node): ?Node
    {
        if (! $this->doctrineEntityManipulator->isStandaloneDoctrineEntityClass($node)) {
            return null;
        }

        $unusedMethodNames = $this->classUnusedPrivateClassMethodResolver->getClassUnusedMethodNames($node);
        if ($unusedMethodNames !== []) {
            $node = $this->removeClassMethodsByNames($node, $unusedMethodNames);
        }

        $unusedPropertyNames = $this->resolveUnusedPrivatePropertyNames($node);
        if ($unusedPropertyNames !== []) {
            $node = $this->removeClassPrivatePropertiesByNames($node, $unusedPropertyNames);
        }

        return $node;
    }

    /**
     * Remove unused methods immediately, so we can then remove unused properties.
     * @param string[] $unusedMethodNames
     */
    private function removeClassMethodsByNames(Class_ $class, array $unusedMethodNames): Class_
    {
        foreach ($class->stmts as $key => $classStmt) {
            if (! $classStmt instanceof ClassMethod) {
                continue;
            }

            if ($this->isNames($classStmt, $unusedMethodNames)) {
                // remove immediately
                unset($class->stmts[$key]);
            }
        }

        return $class;
    }

    /**
     * @return string[]
     */
    private function resolveUnusedPrivatePropertyNames(Class_ $class): array
    {
        $privatePropertyNames = $this->classManipulator->getPrivatePropertyNames($class);

        // get list of fetched properties
        $usedPropertyNames = $this->resolveClassUsedPropertyFetchNames($class);

        return array_diff($privatePropertyNames, $usedPropertyNames);
    }

    /**
     * @param string[] $unusedPropertyNames
     */
    private function removeClassPrivatePropertiesByNames(Class_ $node, array $unusedPropertyNames): Class_
    {
        foreach ($node->stmts as $key => $stmt) {
            if (! $stmt instanceof Property) {
                continue;
            }

            if (! $this->isNames($stmt, $unusedPropertyNames)) {
                continue;
            }

            unset($node->stmts[$key]);

            // remove "$this->someProperty = new ArrayCollection()"
            $propertyName = $this->getName($stmt);
            if (isset($this->collectionByPropertyName[$propertyName])) {
                $this->removeNode($this->collectionByPropertyName[$propertyName]);
            }

            $this->removeInversedByOrMappedByOnRelatedProperty($stmt);
        }

        return $node;
    }

    private function getOtherRelationProperty(Property $property): ?Property
    {
        $targetClass = $this->doctrineEntityManipulator->resolveTargetClass($property);
        $otherProperty = $this->doctrineEntityManipulator->resolveOtherProperty($property);

        if ($targetClass === null || $otherProperty === null) {
            return null;
        }

        // get the class property and remove "mappedBy/inversedBy" from annotation
        $relatedEntityClass = $this->parsedNodesByType->findClass($targetClass);
        if (! $relatedEntityClass instanceof Class_) {
            return null;
        }

        foreach ($relatedEntityClass->stmts as $relatedEntityClassStmt) {
            if (! $relatedEntityClassStmt instanceof Property) {
                continue;
            }

            if (! $this->isName($relatedEntityClassStmt, $otherProperty)) {
                continue;
            }

            return $relatedEntityClassStmt;
        }

        return null;
    }

    private function removeInversedByOrMappedByOnRelatedProperty(Property $property): void
    {
        $otherRelationProperty = $this->getOtherRelationProperty($property);
        if ($otherRelationProperty === null) {
            return;
        }

        $this->doctrineEntityManipulator->removeMappedByOrInversedByFromProperty($otherRelationProperty);
    }

    private function isPropertyFetchAssignOfArrayCollection(PropertyFetch $propertyFetch): bool
    {
        $parentNode = $propertyFetch->getAttribute(AttributeKey::PARENT_NODE);
        if (! $parentNode instanceof Assign) {
            return false;
        }

        if (! $parentNode->expr instanceof New_) {
            return false;
        }

        /** @var New_ $new */
        $new = $parentNode->expr;

        return $this->isName($new->class, 'Doctrine\Common\Collections\ArrayCollection');
    }

    /**
     * @return string[]
     */
    private function resolveClassUsedPropertyFetchNames(Class_ $class): array
    {
        $usedPropertyNames = [];

        $this->traverseNodesWithCallable($class->stmts, function (Node $node) use (&$usedPropertyNames) {
            if (! $node instanceof PropertyFetch) {
                return null;
            }

            if (! $this->isName($node->var, 'this')) {
                return null;
            }

            /** @var string $propertyName */
            $propertyName = $this->getName($node->name);

            // skip collection initialization, e.g. "$this->someProperty = new ArrayCollection();"
            if ($this->isPropertyFetchAssignOfArrayCollection($node)) {
                /** @var Assign $parentNode */
                $parentNode = $node->getAttribute(AttributeKey::PARENT_NODE);
                $this->collectionByPropertyName[$propertyName] = $parentNode;
                return null;
            }

            $usedPropertyNames[] = $propertyName;

            return null;
        });

        return $usedPropertyNames;
    }
}
